# -*- coding: utf-8 -*-

"""
Sub-command for Networks.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015 Dropbox, Inc.'


import click
from . import callbacks


# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('network_address', 'Network'),
    ('prefix_length', 'Prefix'),
    # ('site_id': 'Site ID'),
    ('is_ip', 'Is IP?'),
    ('ip_version', 'IP Ver.'),
    ('parent_id', 'Parent ID'),
    ('attributes', 'Attributes'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Network objects.

    A Network resource can represent an IP Network and an IP Address. Working
    with networks is usually done with CIDR notation.

    Networks can have any number of arbitrary attributes as defined below.
    """


# Add
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Network (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-b',
    '--bulk-add',
    metavar='FILENAME',
    help='Bulk add Networks from the specified colon-delimited file.',
    type=click.File('rb'),
    callback=callbacks.process_bulk_add,
)
@click.option(
    '-c',
    '--cidr',
    metavar='CIDR',
    #help='A network or IP address in CIDR notation.',
    #required=True,
    help='A network or IP address in CIDR notation.  [required]',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Network is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def add(ctx, attributes, bulk_add, cidr, site_id):
    """
    Add a new Network.

    You must provide a Site ID using the -s/--site-id option.

    When adding a new Network, you must provide a value for the -c/--cidr
    option.

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.
    """
    data = bulk_add or ctx.params

    # Enforce required options
    if bulk_add is None:
        if cidr is None:
            raise click.UsageError('Missing option "-c" / "--cidr"')
    ctx.obj.add(data)


# List
@cli.group(invoke_without_command=True)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Network being retrieved.',
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    help='Limit result to N resources.',
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    help='Unique ID of the Site this Network is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def list(ctx, id, limit, offset, site_id):
    """
    List existing Networks for a Site.

    You must provide a Site ID using the -s/--site-id option.

    When listing Networks, all objects are displayed by default. You optionally
    may lookup a single Network by ID using the -i/--id option.

    You may limit the number of results using the -l/--limit option.
    """
    data = ctx.params

    # If we aren't passing a sub-command, just call list(), otherwise let it
    # fallback to default behavior.
    if ctx.invoked_subcommand is None:
        ctx.obj.list(data, display_fields=DISPLAY_FIELDS)


@list.command()
@click.option(
    '-d',
    '--direct',
    is_flag=True,
    help='Return only direct subnets.',
    default=None,
)
@click.option(
    '--include-networks/--no-include-networks',
    is_flag=True,
    help='Include non-IP networks.',
    default=None,
)
@click.option(
    '--include-ips/--no-include-ips',
    is_flag=True,
    help='Include IP addresses.',
    default=None,
)
@click.pass_context
def subnets(ctx, *args, **kwargs):
    """Get subnets of a Network."""
    callbacks.list_endpoint(ctx, display_fields=DISPLAY_FIELDS)


@list.command()
@click.option(
    '-d',
    '--direct',
    is_flag=True,
    help='Return only direct supernets.',
    default=None,
)
@click.pass_context
def supernets(ctx, *args, **kwargs):
    """Get supernets of a Network."""
    callbacks.list_endpoint(ctx, display_fields=DISPLAY_FIELDS)


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Network being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Network is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove a Network.

    You must provide a Site ID using the -s/--site-id option.

    When removing a Network, you must provide the unique ID using -i/--id. You
    may retrieve the ID for a Network by parsing it from the list of Networks
    for a given Site:

        nsot networks list --site-id <site_id> | grep <network>
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this Network (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Network being updated.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Network is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def update(ctx, attributes, id, site_id):
    """
    Update a Network.

    You must provide a Site ID using the -s/--site-id option.

    When updating a Network you must provide the unique ID (-i/--id) and at
    least one of the optional arguments.
    """
    if not any([attributes]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)

    data = ctx.params
    ctx.obj.update(data)



# -*- coding: utf-8 -*-

"""
Base CLI commands for all objects. Model-specific objects and argument parsers
will be defined in subclasses or by way of factory methods.
"""

__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015 Dropbox, Inc.'


import click
import datetime
import logging
import os
import prettytable
from slumber.exceptions import HttpClientError
import sys

import pynsot
from . import client, dotfile
from .models import ApiModel


# Constants/Globals
if os.getenv('DEBUG'):
    logging.basicConfig(level=logging.DEBUG)
log = logging.getLogger(__name__)

# Make the --help option also have -h
CONTEXT_SETTINGS = {
    'help_option_names': ['-h', '--help'],
}

# Where to find the command plugins.
CMD_FOLDER = os.path.abspath(os.path.join(
                             os.path.dirname(__file__), 'commands'))


class NsotCLI(click.MultiCommand):
    """
    Base command object used to define object-specific command-line parsers.

    This will load command plugins from the "commands" folder.

    Plugins must be named "cmd_{foo}.py" and must have a top-level command
    named "cli".
    """
    def list_commands(self, ctx):
        """List all commands from python modules in plugin folder."""
        rv = []
        for filename in os.listdir(CMD_FOLDER):
            if filename.endswith('.py') and filename.startswith('cmd_'):
                rv.append(filename[4:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Import a command module and return it."""
        try:
            if sys.version_info[0] == 2:
                name = name.encode('ascii', 'replace')
            mod = __import__('pynsot.commands.cmd_' + name,
                             None, None, ['app'])
        except ImportError as err:
            print err
            return None
        return mod.cli  # Each cmd_ plugin defines top-level "cli" command


class App(object):
    """Context object for holding state data for the CLI app."""
    def __init__(self, client_args, ctx, verbose=False):
        self.client_args = client_args
        self.ctx = ctx
        self.verbose = verbose
        self.resource_name = self.ctx.invoked_subcommand

    def get_api_client(self, auth_method=None, url=None, email=None,
                       secret_key=None, default_site=None):
        """
        Safely create an API client so that users don't see tracebacks.

        :param auth_method:
            Auth method used by the client

        :param url:
            API URL

        :param email:
            User's email

        :param secret_key:
            User's secret_key

        :param default_site:
            User's default site_id
        """
        try:
            client_class, arg_names = client.get_auth_client_info(auth_method)
        except KeyError:
            raise click.UsageError('Invalid auth_method: %s' % (auth_method,))

        # Construct kwargs to pass to the client_class
        local_vars = locals()
        kwargs = {arg_name: local_vars[arg_name] for arg_name in arg_names}
        log.debug('APP KWARGS = %r' % (kwargs,))
        try:
            api_client = client_class(url, **kwargs)
        except client.ClientError as err:
            msg = str(err)
            if 'Connection refused' in msg:
                msg = 'Could not connect to server: %s' % (url,)
            raise click.UsageError(msg)
        return api_client

    @property
    def api(self):
        """This way the API client is not created until called."""
        if not hasattr(self, '_api'):
            self._api = self.get_api_client(**self.client_args)
        return self._api

    @property
    def singular(self):
        """Return singular form of resource_name. (e.g. "sites" -> "site")"""
        resource_name = self.resource_name
        if resource_name.endswith('s'):
            resource_name = resource_name[:-1]
        return resource_name

    @property
    def resource(self):
        """
        Return an API resource method for calling endpoints.

        For example if ``resource_name`` is ``networks``, this is equivalent to
        calling ``self.api.networks``.
        """
        return self.api.get_resource(self.resource_name)

    @staticmethod
    def pretty_dict(data, delim='=', sep=', '):
        """
        Return a dict in k=v format.

        :param dict:
            A dict

        :param sep:
            Character used to separate items
        """
        pretty = sep.join(
            '%s%s%s' % (k, delim, v) for k, v in
            data.iteritems()
        )
        return pretty

    def format_message(self, obj_single, message=''):
        """
        Attempt to make messages human-readable.

        :param obj_single:
            Singular object name

        :param message:
            Dat message tho!
        """
        log.debug('FORMATTING MESSAGE: %r' % (message,))
        if 'UNIQUE constraint failed' in message:
            message = '%s object already exists.' % (obj_single.title(),)
        return message

    def format_timestamp(self, ts):
        """
        Take a UNIX timestamp and make it a datetime string.

        :param ts:
            UNIX timestamp
        """
        ts = int(ts)
        dt = datetime.datetime.fromtimestamp(ts)
        return str(dt)

    def handle_error(self, action, data, err):
        """
        Handle error API response.

        :param action:
            The action name

        :param data:
            Dict of arguments

        :param err:
            Exception object
        """
        resp = getattr(err, 'response', None)
        obj_single = self.singular
        extra = '\n'

        # If it's an API error, format it all pretty-like for human eyes.
        if resp is not None:
            body = resp.json()
            log.debug('API ERROR: %r' % (body,))

            msg = body['error']['message']
            msg = self.format_message(obj_single, msg)

            # Add the status code and reason to the output
            log.debug('ERROR MESSAGE = %r' % (msg,))
            if self.verbose or not msg:
                t_ = '%s %s'
                reason = resp.reason.upper()
                extra += t_ % (resp.status_code, reason)
            if not msg:
                msg = extra.strip()
        else:
            msg = str(err)

        # If we're being verbose, print some extra context.
        if self.verbose:
            t_ = ' trying to %s %s with args: %s'
            pretty_dict = self.pretty_dict(data)
            extra += t_ % (action, obj_single, pretty_dict)
            msg += extra

        # Colorize the failure text as red.
        self.ctx.exit(click.style('[FAILURE] ', fg='red') + msg)

    def handle_response(self, action, data, result):
        """
        Handle positive API response.

        :param action:
            The action name

        :param data:
            Dict of arguments

        :param result:
            Dict containing result
        """
        pretty_dict = self.pretty_dict(data)
        t_ = '%s %s with args: %s!'
        if action.endswith('e'):
            action = action[:-1]  # "remove" -> "remov"
        action = action.title() + 'ed'  # "remove" -> "removed"
        msg = t_ % (action, self.singular,  pretty_dict)

        # Colorize the success text as green.
        click.echo(click.style('[SUCCESS] ', fg='green') + msg)

    def map_fields(self, fields, fields_map):
        """
        Map ``fields`` using ``fields_map`` for table display.

        :param fields:
            List of field names

        :param fields_map:
            Mapping of field names to translations
        """
        log.debug('MAP_FIELDS FIELDS = %r' % (fields,))
        log.debug('MAP_FIELDS FIELDS_MAP = %r' % (fields_map,))
        try:
            headers = [fields_map[f] for f in fields]
        except KeyError as err:
            msg = 'Could not map field %s when displaying results.' % (err,)
            self.ctx.exit(msg)
        log.debug('MAP_FIELDS HEADERS = %r' % (headers,))
        return headers

    def format_field(self, field, field_data):
        """
        Specially format a field.

        :param field:
            Field name

        :param field_data:
            Field data dict
        """
        # If it's a user field, only show the email
        if field == 'user':
            field_data = field_data['email']

        # If the field is a dict, pretty_dict it!
        if isinstance(field_data, dict):
            # If this is an inner dict, prettify it, too.
            for sub_field in ('attributes', 'constraints'):
                if sub_field in field_data:
                    field_data[sub_field] = self.pretty_dict(
                        field_data[sub_field], sep='\n')

            # This is so that k=v looks better when printing a resource's
            # contents
            if field == 'resource':
                delim = ':'
            # Otherwise we just fallback to standard k=v
            else:
                delim = '='
            field_data = self.pretty_dict(field_data, delim=delim, sep='\n')

        # If it's the 'change_at' field, make it human-readable
        elif field == 'change_at':  # Timestamp
            field_data = self.format_timestamp(field_data)

        return field_data

    def print_list(self, objects, display_fields):
        """
        Print a list of objects in a table format.

        :param objects:
            List of object dicts

        :param display_fields:
            Ordered list of 2-tuples of (field, display_name) used
            to translate field names for display
        """
        # Extract the field names and create a mapping used for translation
        fields = [f[0] for f in display_fields]  # Field names are 1st item
        fields_map = dict(display_fields)

        # Human-readable field headings as they will be displayed
        headers = self.map_fields(fields, fields_map)

        # Order the object key/val by the order in display fields
        table_data = []

        # We're doing all of this just so we can pretty print dicts as k=v
        for obj in objects:
            obj_list = []
            for field in fields:
                field_data = obj[field]

                # Attempt to format the field
                field_data = self.format_field(field, field_data)

                obj_list.append(field_data)
            table_data.append(obj_list)

        # Prepare the table object
        table = prettytable.PrettyTable(headers)
        table.vrules = prettytable.FRAME  # Display table in a frame
        table.align = 'l'  # Left-align everything
        for row in table_data:
            table.add_row(row)

        # Only paginate if table is longer than terminal.
        _, t_height, = click.get_terminal_size()
        if len(table_data) > t_height:
            click.echo_via_pager(table)
        else:
            click.echo(table)

    def rebase(self, data):
        """
        If this is not a site object, then rebase the API URL.

        :param data:
            Dict of query arguments
        """
        # Prefer site_id from args, or default to API client's default_site
        # provided in user's config.
        site_id = data.pop('site_id', None)
        if site_id is None and self.resource_name != 'sites':
            site_id = self.api.default_site

        log.debug('Got site_id: %s' % site_id)
        if site_id is not None:
            log.debug('Site_id found; rebasing API URL!')
            self.api._store['base_url'] += '/sites/' + site_id

    def add(self, data):
        """POST"""
        action = 'add'
        log.debug('adding %s' % data)
        self.rebase(data)

        try:
            result = self.resource.post(data)
        except HttpClientError as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)

    def list(self, data, display_fields=None, resource=None):
        """GET"""
        action = 'list'
        log.debug('listing %s' % data)
        obj_id = data.get('id')  # If obj_id, it's a single object

        # If a resource object is provided, call it instead, and only rebase if
        # we haven't provided our own resource.
        if resource is None:
            self.rebase(data)  # Rebase first
            resource = self.resource

        try:
            # Try getting a single object first
            if obj_id:
                result = resource(obj_id).get()
            # Or get all of them.
            else:
                result = resource.get(**data)
        except HttpClientError as err:
            self.handle_error(action, data, err)
        else:
            objects = []
            # Turn a single object into a list
            if obj_id:
                obj = result['data'][self.singular]
                objects = [obj]
            # Or just list all of them.
            elif result:
                objects = result['data'][self.resource_name]

            if objects:
                self.print_list(objects, display_fields)
            else:
                pretty_dict = self.pretty_dict(data)
                t_ = 'No %s found matching args: %s!'
                msg = t_ % (self.singular, pretty_dict)
                click.echo(msg)

    def remove(self, **data):
        """DELETE"""
        action = 'remove'
        obj_id = data['id']
        log.debug('removing %s' % obj_id)
        self.rebase(data)

        try:
            result = self.resource(obj_id).delete()
        except HttpClientError as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)

    def update(self, data):
        """PUT"""
        action = 'update'
        obj_id = data.pop('id')
        log.debug('updating %s' % data)
        self.rebase(data)

        # Get the original object by id first so that we can keep any existing
        # values without resetting them.
        try:
            obj = self.resource(obj_id).get()
        except HttpClientError as err:
            self.handle_error(action, data, err)
        else:
            model = ApiModel(obj)
            payload = dict(model)
            payload.pop('id')  # We don't want id when doing a PUT

        # Update the payload from the CLI params if the value isn't null.
        for k, v in data.iteritems():
            if v is not None:
                payload[k] = v

        # And now we call PUT
        try:
            result = self.resource(obj_id).put(payload)
        except HttpClientError as err:
            self.handle_error(action, data, err)
        else:
            self.handle_response(action, data, result)


@click.command(cls=NsotCLI, context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--verbose', is_flag=True, help='Toggle verbosity.')
@click.version_option(version=pynsot.__version__)
@click.pass_context
def app(ctx, verbose):
    """
    Network Source of Truth (NSoT) command-line utility.

    For detailed documentation, please visit https://nsot.readthedocs.org
    """
    # This is the "app" object attached to all contexts.

    # Read the dotfile
    try:
        client_args = dotfile.Dotfile().read()
    except dotfile.DotfileError as err:
        raise click.UsageError(err.message)

    # Construct the App!
    ctx.obj = App(client_args=client_args, ctx=ctx, verbose=verbose)

    # Store the invoked_subcommand (e.g. 'networks') name as parent_name so that
    # descendent sub-commands can reference where they came from, such as when
    # calling callbacks.list_endpoint()
    ctx.obj.parent_name = ctx.invoked_subcommand

if __name__ == '__main__':
    app()



# -*- coding: utf-8 -*-

"""
Simple Python API client for NSoT REST API
"""

__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015 Dropbox, Inc.'


import json
import logging
from requests.auth import AuthBase
import slumber
from slumber.exceptions import HttpClientError


# Logger
log = logging.getLogger(__name__)

# Header used for passthrough authentication.
AUTH_HEADER = 'X-NSoT-Email'


class ClientError(HttpClientError):
    """Generic client error."""


class LoginFailed(ClientError):
    """Raised when login fails for some reason."""


class BaseClient(slumber.API):
    """
    Magic REST API client for NSoT.
    """
    def __init__(self, base_url=None, **kwargs):
        auth = kwargs.pop('auth', None)  # Ditch default auth
        self.default_site = kwargs.pop('default_site', None)  # Default site_id

        # Override the auth method if we have defined .get_auth()
        if auth is None:
            # Set these as object attributes so that they can be mutated in the
            # subclass .get_auth() method
            self._kwargs = kwargs
            auth = self.get_auth(base_url)

        kwargs['auth'] = auth
        kwargs['append_slash'] = False  # No slashes!
        super(BaseClient, self).__init__(base_url, **kwargs)

    def get_auth(self, base_url=None):
        """
        Subclasses should references kwargs from ``self._kwargs``.

        :param base_url:
            Base API URL
        """
        raise NotImplementedError('SUBCLASS ME OR SOMETHING!')

    def error(self, exc):
        """
        Take errors and make them human-readable.

        :param exc:
            Exception instance
        """
        log.debug('Processing error: %r' % (exc,))
        # If it's a HTTP response, format the JSON
        try:
            err = exc.response.json()['error']
            msg = '%s %s' % (err['code'], err['message'])
        except AttributeError:
            msg = str(exc)
        raise ClientError(msg)

    def get_resource(self, resource_name):
        """
        Return a single resource object.

        :param resource_name:
            Name of resource
        """
        return getattr(self, resource_name)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '<%s(url=%s)>' % (cls_name, self._store['base_url'])


class EmailHeaderAuthentication(AuthBase):
    """Special authentication that sets the email auth header."""
    def __init__(self, email=None, auth_header=AUTH_HEADER):
        if email is None:
            raise LoginFailed('You must provide an email!')
        self.email = email
        self.auth_header = auth_header

    def __call__(self, r):
        """Set the auth header."""
        r.headers[self.auth_header] = self.email
        return r


class EmailHeaderClient(BaseClient):
    """Default client using email auth header method."""
    def get_auth(self, base_url):
        kwargs = self._kwargs
        email = kwargs.pop('email', None)
        auth_header = kwargs.pop('auth_header', AUTH_HEADER)
        return EmailHeaderAuthentication(email, auth_header)


class AuthTokenAuthentication(AuthBase):
    """
    Special authentication that utilizes auth_tokens.

    Adds header for "Authorization: ApiToken {email}:{auth_token}"
    """
    def __init__(self, email=None, auth_token=None):
        self.email = email
        self.auth_token = auth_token

    def __call__(self, r):
        header = 'AuthToken %s:%s' % (self.email, self.auth_token)
        r.headers['Authorization'] = header
        return r


class AuthTokenClient(BaseClient):
    """Client that uses auth_token method."""
    def get_token(self, base_url, email, secret_key):
        """
        Currently ghetto: Hit the API to get an auth_token.

        :param base_url:
            API URL

        :param email:
            User's email

        :param secret_key:
            User's secret_key
        """
        data = {'email': email, 'secret_key': secret_key}
        log.debug('Getting token for user data: %r' % (data,))
        try:
            url = base_url + '/authenticate'
            headers = {'content-type': 'application/json'}
            r = slumber.requests.post(url, data=json.dumps(data),
                                      headers=headers)
        except Exception as err:
            log.debug('Got error: %s' % (err,))
            self.error(err)

        if r.ok:
            log.debug('Got response: %r' % (r,))
            return r.json()['data']['auth_token']
        else:
            msg = 'Failed to fetch auth_token'
            err = HttpClientError(msg, response=r, content=r.content)
            self.error(err)

    def get_auth(self, base_url):
        kwargs = self._kwargs
        email = kwargs.pop('email', None)
        secret_key = kwargs.pop('secret_key', None)
        auth_token = self.get_token(base_url, email, secret_key)
        auth = AuthTokenAuthentication(email, auth_token)
        return auth
Client = AuthTokenClient  # Default client is auth_token


# Mapping to our two (2) hard-coded auth methods and their required arguments.
AUTH_CLIENTS = {
    'auth_header': (EmailHeaderClient, ('email', 'default_site')),
    'auth_token': (AuthTokenClient, ('email', 'secret_key', 'default_site')),
}


def get_auth_client_info(auth_method):
    """
    Return the proper Client class and required args.

    :param auth_method:
        Auth method used by the client
    """
    return AUTH_CLIENTS[auth_method]



"""
Handle the read, write, and generation of the .pynsotrc config file.
"""

__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015 Dropbox, Inc.'


from ConfigParser import RawConfigParser
import click
import logging
import os
import rcfile
import stat


# Logging object
log = logging.getLogger(__name__)

# Default authentication method (one of user_auth_header, auth_token)
AUTH_METHOD = 'auth_token'

# Mapping of required field names and default values we want to be in the
# dotfile.
REQUIRED_FIELDS = {
    'auth_method': AUTH_METHOD,
    'url': None,
    'email': None,
    'secret_key': None,
}

# Path stuff
USER_HOME = os.path.expanduser('~')
DOTFILE_NAME = '.pynsotrc'
DOTFILE_PATH = os.path.join(USER_HOME, DOTFILE_NAME)
DOTFILE_PERMS = 0600  # -rw-------

# Config section name
SECTION_NAME = 'pynsot'


class DotfileError(Exception):
    """Raised when something with the dotfile fails."""


class Dotfile(object):
    """Create, read, and write a dotfile."""
    def __init__(self, filepath=DOTFILE_PATH, **kwargs):
        self.filepath = filepath

    def read(self, **kwargs):
        """
        Read ``~/.pynsotrc`` and return it as a dict.
        """
        # If the dotfile isn't found, prompt the user to create one.
        if not os.path.exists(self.filepath):
            p = '%s not found; would you like to create it?' % (self.filepath,)
            if click.confirm(p, default=True, abort=True):
                config_data = self.get_config_data(**kwargs)
                self.write(config_data)

        # Validate the permissions and presence of fields in the dotfile
        self.validate_perms()
        config = rcfile.rcfile(SECTION_NAME, args={'config': self.filepath})
        self.validate_fields(config)

        config.pop('config', None)  # We don't need this field in here.
        self.config = config
        return config

    def validate_perms(self):
        """Make sure dotfile ownership and permissions are correct."""
        # Ownership
        s = os.stat(self.filepath)
        if s.st_uid != os.getuid():
            msg = '%s: %s must be owned by you' % (DOTFILE_NAME, self.filepath)
            raise DotfileError(msg)

        # Permissions
        self.enforce_perms()

    def validate_fields(self, field_names):
        """
        Make sure all the fields are set.

        :param field_names:
            List of field names to validate
        """
        for rname in sorted(REQUIRED_FIELDS):
            if rname not in field_names:
                msg = '%s: Missing required field: %s' % (DOTFILE_NAME, rname)
                raise DotfileError(msg)

    def enforce_perms(self, perms=DOTFILE_PERMS):
        """
        Enforce permissions on the dotfile.

        :param perms:
            Octal number representing permissions to enforce
        """
        log.debug('Enforcing permissions %o on %s' % (perms, self.filepath))
        os.chmod(self.filepath, perms)

    def write(self, config_data, filepath=None):
        """
        Create a dotfile from keyword arguments.

        :param config_data:
            Dict of config settings

        :param filepath:
            (Optional) Path to write
        """
        if filepath is None:
            filepath = self.filepath
        config = RawConfigParser()
        section = SECTION_NAME
        config.add_section(section)

        # Set the config settings
        for key, val in config_data.iteritems():
            config.set(section, key, val)

        with open(filepath, 'wb') as dotfile:
            config.write(dotfile)

        self.enforce_perms()
        log.debug('wrote %s' % filepath)

    @staticmethod
    def get_config_data(required_fields=REQUIRED_FIELDS, **kwargs):
        """
        Enumerate required fields and prompt for ones that weren't provided if
        they don't have default values.

        :param required_fields:
            Mapping of required field names to default values

        :param kwargs:
            Dict of config settings
        """
        config_data = {}
        for field, default_value in required_fields.iteritems():
            # If the field was not provided and does not have a default value
            # prompt for it
            if field not in kwargs and default_value is None:
                prompt = 'Please enter %s' % (field.upper(),)
                value = click.prompt(prompt, type=str)
            # Or, use the provided value
            elif field in kwargs:
                value = kwargs[field]
            # Otherwise, use the default value
            else:
                value = default_value

            config_data[field] = value
        return config_data



"""
Callbacks used in handling command plugins.
"""

import ast
import click
import csv
import logging


log = logging.getLogger(__name__)

# Objects that do not have attribtues
NO_ATTRIBUTES = ('attributes',)


def process_site_id(ctx, param, value):
    """
    Callback to attempt to get site_id from ``~/.pynsotrc`` if it's not
    provided using -s/--site-id.
    """
    log.debug('GOT DEFAULT_SITE: %s' % ctx.obj.api.default_site)
    log.debug('GOT PROVIDED SITE_ID: %s' % value)

    # Try to get site_id from the app config, or complain that it's not set.
    if value is None:
        default_site = ctx.obj.api.default_site
        if default_site is None:
            raise click.UsageError('Missing option "-s" / "--site-id".')
        value = default_site

    return value


def transform_attributes(ctx, param, value):
    """Callback to turn attributes arguments into a dict."""
    attrs = {}

    # If value is a string, we'll assume that it's comma-separated.
    if isinstance(value, basestring):
        value = value.split(',')

    for attr in value:
        key, _, val = attr.partition('=')
        if not all([key, val]):
            msg = 'Invalid attribute: %s; format should be key=value' % (attr,)
            raise click.UsageError(msg)
        attrs[key] = val
    return attrs


def transform_event(ctx, param, value):
    """Callback to transform event into title case."""
    if value is not None:
        return value.title()
    return value


def transform_resource_name(ctx, param, value):
    """Callback to transform resource_name into title case."""
    if value is not None:
        return value.title()
    return value


def process_bulk_add(ctx, param, value):
    """
    Callback to parse bulk addition of objects from a colon-delimited file.

    Format:

    + The first line of the file must be the field names.
    + Attribute pairs must be commma-separated, and in format k=v
    + The attributes must exist!
    """
    if value is None:
        return value

    # This is our object name (e.g. 'devices')
    parent_name = ctx.obj.parent_name
    objects = []

    # Value is already an open file handle
    reader = csv.DictReader(value, delimiter=':')
    for r in reader:
        lineno = reader.line_num

        # Make sure the file is correctly formatted.
        if len(r) != len(reader.fieldnames):
            msg = 'File has wrong number of fields on line %d' % (lineno,)
            raise click.BadParameter(msg)

        # Transform attributes for eligible resource types
        if parent_name not in NO_ATTRIBUTES:
            attributes = transform_attributes(
                ctx, param, r['attributes']
            )
            r['attributes'] = attributes

        # Transform True, False into booleans
        for k, v in r.iteritems():
            if not isinstance(v, basestring):
                msg = 'Error parsing file on line %d' % (lineno,)
                raise click.BadParameter(msg)
            if v.title() in ('True', 'False'):
                r[k] = ast.literal_eval(v)
        objects.append(r)

    log.debug('PARSED BULK DATA = %r' % (objects,))

    # Return a dict keyed by calling object name
    return {parent_name: objects}


def list_endpoint(ctx, display_fields):
    """
    Determine params and a resource object to pass to ``ctx.obj.list()``

    :param ctx:
        Context from the calling command

    :param display_fields:
        Display fields used to list object results.
    """
    # Gather our args from our parent and ourself
    data = ctx.parent.params
    data.update(ctx.params)

    parent_id = data.pop('id')
    site_id = data.pop('site_id')

    # Make sure that parent_id is provided.
    if parent_id is None:
        raise click.UsageError('Missing option "-i" / "--id".')

    # Use our name, parent's command name, and the API object to retrieve the
    # endpoint resource used to call this endpoint.
    api = ctx.obj.api
    parent_name = ctx.obj.parent_name  # e.g. 'networks'
    my_name = ctx.info_name  # e.g. 'supernets'

    # e.g. /api/sites/1/networks/5/supernets
    parent_resource = getattr(api.sites(site_id), parent_name)
    my_resource = getattr(parent_resource(parent_id), my_name)

    ctx.obj.list(data, display_fields=display_fields, resource=my_resource)



# -*- coding: utf-8 -*-

"""
Sub-command for Devices.

In all cases ``data = ctx.params`` when calling the appropriate action method
on ``ctx.obj``. (e.g. ``ctx.obj.add(ctx.params)``)

Also, ``action = ctx.info_name`` *might* reliably contain the name of the
action function, but still not sure about that. If so, every function could be
fundamentally simplified to this::

    getattr(ctx.obj, ctx.info_name)(ctx.params)
"""

__author__ = 'Jathan McCollum'
__maintainer__ = 'Jathan McCollum'
__email__ = 'jathan@dropbox.com'
__copyright__ = 'Copyright (c) 2015 Dropbox, Inc.'


import click
from . import callbacks


# Ordered list of 2-tuples of (field, display_name) used to translate object
# field names oto their human-readable form when calling .print_list().
DISPLAY_FIELDS = (
    ('id', 'ID'),
    ('hostname', 'Hostname'),
    # ('site_id': 'Site ID'),
    ('attributes', 'Attributes'),
)


# Main group
@click.group()
@click.pass_context
def cli(ctx):
    """
    Device objects.

    A device represents various hardware components on your network such as
    routers, switches, console servers, PDUs, servers, etc.

    Devices also support arbitrary attributes similar to Networks.
    """


# Add
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this network (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-b',
    '--bulk-add',
    metavar='FILENAME',
    help='Bulk add Devices from the specified colon-delimited file.',
    type=click.File('rb'),
    callback=callbacks.process_bulk_add,
)
@click.option(
    '-H',
    '--hostname',
    metavar='HOSTNAME',
    #help='The hostname of the Device.',
    #required=True,  # What to do about required flags!
    help='The hostname of the Device.  [required]',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def add(ctx, attributes, bulk_add, hostname, site_id):
    """
    Add a new Device.

    You must provide a Site ID using the -s/--site-id option.

    When adding a new Device, you must provide a value for the -h/--hostname
    option.

    If you wish to add attributes, you may specify the -a/--attributes
    option once for each key/value pair.
    """
    data = bulk_add or ctx.params

    # Enforce required options
    if bulk_add is None:
        if hostname is None:
            raise click.UsageError('Missing option "-H" / "--hostname".')

    ctx.obj.add(data)


# List
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Device being retrieved.',
)
@click.option(
    '-l',
    '--limit',
    metavar='LIMIT',
    type=int,
    help='Limit result to N resources.',
)
@click.option(
    '-o',
    '--offset',
    metavar='OFFSET',
    help='Skip the first N resources.',
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def list(ctx, id, limit, offset, site_id):
    """
    List existing Devices for a Site.

    You must provide a Site ID using the -s/--site-id option.

    When listing Devices, all objects are displayed by default. You may
    optionally lookup a single Device by ID using the -i/--id option.

    You may limit the number of results using the -l/--limit option.
    """
    data = ctx.params
    ctx.obj.list(data, display_fields=DISPLAY_FIELDS)


# Remove
@cli.command()
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Device being deleted.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def remove(ctx, id, site_id):
    """
    Remove a Device.

    You must provide a Site ID using the -s/--site-id option.

    When removing a Device, you must provide the unique ID using -i/--id. You
    may retrieve the ID for a Device by parsing it from the list of Devices
    for a given Site:

        nsot devices list --site-id <site_id> | grep <hostname>
    """
    data = ctx.params
    ctx.obj.remove(**data)


# Update
@cli.command()
@click.option(
    '-a',
    '--attributes',
    metavar='ATTRS',
    help='A key/value pair attached to this network (format: key=value).',
    multiple=True,
    callback=callbacks.transform_attributes,
)
@click.option(
    '-H',
    '--hostname',
    metavar='HOSTNAME',
    help='The hostname of the Device.',
)
@click.option(
    '-i',
    '--id',
    metavar='ID',
    type=int,
    help='Unique ID of the Device being updated.',
    required=True,
)
@click.option(
    '-s',
    '--site-id',
    metavar='SITE_ID',
    type=int,
    help='Unique ID of the Site this Device is under.  [required]',
    callback=callbacks.process_site_id,
)
@click.pass_context
def update(ctx, attributes, hostname, id, site_id):
    """
    Update a Device.

    You must provide a Site ID using the -s/--site-id option.

    When updating a Device you must provide the unique ID (-i/--id) and at
    least one of the optional arguments.
    """
    if not any([attributes, hostname]):
        msg = 'You must supply at least one of the optional arguments.'
        raise click.UsageError(msg)

    data = ctx.params
    ctx.obj.update(data)
